import pyeda
import re
from pyeda.inter import *

pyeda.boolalg.expr.Variable = pyeda.boolalg.expr.ExprVariable
pyeda.boolalg.expr._One = pyeda.boolalg.expr._ExprOne
pyeda.boolalg.expr._Zero = pyeda.boolalg.expr._ExprZero
pyeda.boolalg.expr.Complement = pyeda.boolalg.expr.ExprComplement
pyeda.boolalg.expr.NotOp = pyeda.boolalg.expr.ExprNot
pyeda.boolalg.expr.OrOp = pyeda.boolalg.expr.ExprOr
pyeda.boolalg.expr.AndOp = pyeda.boolalg.expr.ExprAnd

re_ss = re.compile(r"^stablestate:\s*(.*)")
re_equation = re.compile(r"^equation:\s+(.*?)\s*\*")
re_rule = re.compile(r"^equation.*=\s*(.*)")
re_replace_vars = re.compile(r'\b\w+\b')
re_notation = re.compile(r"(\b\w+)(′)")


import copy
import mpbn
import pandas as pd
import pyeda
from pyeda.inter import *
from colomoto_jupyter import tabulate
from boolevard.utils import *

def Pert(model, perturbation, additive = True):
    '''
    Perturb the model by creating a perturbation node that targets a specific node in the model, simulating a positive (ACT) or negative (INH) effect in the target.
    
    Arguments
    ---------
    model (obj): BooLEV object.
    tNodes (list): list of target nodes to evaluate.
    perturbation (str): string containing the target node and the perturbation type separated by a percentage symbol. E.g. "Node%ACT", "Node%INH".
    additive (bool): if True, the perturbation is additive (i.e. the regulation is incorporated to the target node's rule). Otherwise, the perturbation is substitutive (i.e. the regulation replaces the target node's rule). By default: True    ---------

    Returns
    ---------
    BooLEV object
        Model (BooLEV object) updated with the perturbation.
    ---------
    '''
    pert_model = copy.copy(model)
    content = pert_model._bnet
    rules = {}

    for line in content:
        if "#" not in line and "targets" not in line and "target" not in line:
            node, rule = line.split(", ", 1)
            rules[node] = rule
    
    tnode, pertType = perturbation.split("%")
    new_rules = rules.copy()

    if additive:

        if pertType == "INH":
            new_rules[tnode] = "(" + new_rules[tnode] + ")" + " & !" + tnode + "_" + pertType

        elif pertType == "ACT":
            new_rules[tnode] = "(" + new_rules[tnode] + ")" + " | " + tnode + "_" + pertType
    
    else:

        if pertType == "INH":
            new_rules[tnode] = new_rules[tnode] = "!" + tnode + "_" + pertType

        elif pertType == "ACT":
            new_rules[tnode] = new_rules[tnode] = tnode + "_" + pertType

    pnode = tnode + "_" + pertType
    new_rules[pnode] = pnode
    new_content = []

    new_content.append(f"# Perturbation: {perturbation}")
    new_content.append("targets, factors")

    for idx, (target, factor) in enumerate(new_rules.items()):
        new_content.append(f"{target}, {factor}")

    pert_model._bnet = "\n".join(new_content).splitlines()
    pert_model.Nodes = list(new_rules.keys())
    pert_model.DNFs = dict(zip(pert_model.Nodes, [expr(rule.replace(" ", "").replace("!", "~")).to_dnf() for rule in new_rules.values()]))
    pert_model.NDNFs = dict(zip(pert_model.Nodes, [ExprNot(expr(rule.replace(" ", "").replace("!", "~"))).to_dnf() for rule in new_rules.values()]))
    pert_model._IsPert = any(line.startswith("# Perturbation") for line in pert_model._bnet)
    pert_model._Pert = pert_model._Pert + [line.split(":")[1].strip() for line in pert_model._bnet if line.startswith("# Perturbation")]
    
    mpbn_model = mpbn.MPBooleanNetwork(new_rules)
    pert_model.SS = tabulate(list(mpbn_model.attractors()))
    pert_model.Info = pd.concat([pert_model.SS.transpose(), pd.DataFrame({"DNF": pert_model.DNFs, "NDNF": pert_model.NDNFs}, index = pert_model.Nodes)], axis = 1) 
    pert_model.Info = pert_model.Info.loc[:, ~(pert_model.Info == "*").any()]
    pert_model.Info = pert_model.Info.loc[:, (pert_model.Info.loc[[pnode]] == 1).all(axis = 0) | pert_model.Info.columns.str.contains("DNF")]
    
    return pert_model


import mpbn
import os
import pandas as pd
from pyeda import *
from pyeda.inter import *
from colomoto_jupyter import tabulate
from boolevard.utils import *
from boolevard.transduction import *
from boolevard.perturbations import *

class BooLEV:

    def __init__(self, file_path:str):
        '''
        Initialites the BooLEV object by loading the model from a .bnet file.
        '''
        self._bnet = open(file_path, "r").read().splitlines()   
        self.Info = self._ModelInfo(file_path)
        self.Nodes = list(self.Info.index)
        self.DNFs = dict(zip(self.Nodes, self.Info["DNF"]))
        self.NDNFs = dict(zip(self.Nodes, self.Info["NDNF"]))
        self.SS = self.Info.drop(["DNF", "NDNF"], axis = 1)
        self._IsPert = any(line.startswith("# Perturbed") for line in self._bnet)
        self._Pert = [line.split(":")[1].strip() for line in self._bnet if line.startswith("# Perturbed")]

    def __repr__(self):
        '''
        Returns the representation of the BooLEV object
        '''
        return f"<BooLEV object at {hex(id(self))}>"
    
    # Internal functions
    def _ModelInfo(self, file_path:str):
        '''
        Retrieves the information of the model, including nodes, cDNFs and cNDNFs, and stable states.
        '''
        content = self._bnet
        nodes = []
        DNFs = []
        NDNFs = []

        for line in content:
            if "#" not in line and "targets" not in line and "target" not in line: 
                node, rule = line.split(",")
                nodes.append(node)
                DNFs.append(expr(rule.replace(" ", "").replace("!", "~")).to_dnf())
                NDNFs.append(ExprNot(expr(rule.replace(" ", "").replace("!", "~"))).to_dnf())
        
        model = mpbn.load(file_path)
        SS = tabulate(list(model.attractors()))
        info = pd.concat([SS.transpose(), pd.DataFrame({"DNF": DNFs, "NDNF": NDNFs}, index = nodes)], axis = 1)
        info = info.loc[:, ~(info == "*").any()]
        return info
    
    # Methods
    def Export(self, file_path:str):
        """
        Exports the model in .bnet format.

        Parameters
        **********
        file_path: str
            Path to export the model.

        Returns:
        ********
        BooLEV: object
            The BooLEV object.
        
        Example:
        ********
        >>> model = blv.Load("model.bnet")
        >>> model.Export("model.bnet")
        """

        with open(file_path, "w") as f:
            
            if self._IsPert == True:
                f.write(f"# Perturbed model with {self._Pert} perturbation(s)\n")
            
            f.write("targets, factors\n")

            for line, (target, factor) in enumerate(zip(self.Nodes, self.Info["DNF"])):
                factor = factor.to_unicode()
                factor = re_notation.sub(lambda x: "!" + x.group(1), factor).replace("+", "|").replace("·", "&").replace("¬", "!")
                if line < len(self.Nodes) - 1:
                    f.write(f"{target}, {factor}\n")

                else:
                    f.write(f"{target}, {factor}")
        return self
    
    def Drivers(self, ss: int):
        """
        Extracts drivers from (N)DNFs based on the local state of the node within a given stable state.
        
        Parameters:
        ********
        ss: int 
            Stable State to evaluate.

        Returns:
        ********
        dict
            Dictionary containing de drivers of each node within the stable state.

        Example:
        ********
        >>> model = blv.Load("model.bnet")
        >>> model.Drivers(1) # Compute the drivers for each node within the stable state number 1
        """
        return Drivers(self.Info, ss)

    def CountPaths(self, tNodes:list, ss_wise = False):
        """
        Calculates the signed path count leading to the local state of a node contained in a list. Positive if the local state is 1 and negative if 0.
        
        Parameters:
        **********
        tNodes: list
            List of target nodes to evaluate.
        ss_wise: bool
            If True, returns a list with the corresponding path counts leading to a target's local state for each stable state. Otherwise, it computes the average path count across all stable states contained in the Info attribute of the BooLEV object. By default: False.

        Returns:
        **********
        list
            Signed number of paths leading to the local states of the target nodes. Negative if the local state is 0, positive if 1.
        
        Example:
        **********
        >>> model = blv.Load("model.bnet")
        >>> model.CountPaths(["Node1", "Node2"], ss_wise = False)
        """
        return CountPaths(self.Info, tNodes, ss_wise)
    
    def Pert(self, perturbation:str, additive = True):
        """
        Perturbs the model by creating a perturbation node that targets a specific node in the model, simulating a positive (ACT) or negative (INH) effect in the target.
        
        Parameters:
        ********
        tNodes: list
            List of target nodes to evaluate.
        perturbation: str
            String containing the target node and the perturbation type separated by a percentage symbol. E.g. `"Node%ACT"`, `"Node%INH"`.
        additive: bool
            If True, the perturbation is additive (i.e. the regulation is incorporated to the target node's rule). Otherwise, the perturbation is substitutive (i.e. the regulation replaces the target node's rule). By default: True.

        Returns:
        ********
        BooLEV object
            Model (BooLEV object) updated with the perturbation.
        
        Example:
        ********
        >>> model = blv.Load("model.bnet")
        >>> model.Pert("Node%ACT", additive = True)
        >>> model.Pert("Node%INH", additive = False)
        """
        return Pert(self, perturbation, additive)
    

def Load(file_path:str):
    """
    Loads a model in ``b.net`` format and returns a BooLEV-class object.

    This function loads a Boolean model from a `.bnet` file and returns a 
    `BooLEV` object containing the model's structure and associated data.

    Parameters:
    **********
    file_path : str
        Path to the ``.bnet`` file.

    Returns:
    **********
    BooLEV: object
        A `BooLEV` object containing the model with the following attributes:
        
        - `Nodes` (list): List containing the nodes of the model.
        - `DNFs` (dict): Dictionary with the canonical Disjunctive Normal Form (cDNF) of each node.
        - `NDNFs` (dict): Dictionary with the cDNF of the negated rule of each node.
        - `SS` (pd.DataFrame): DataFrame containing the stable states.
        - `Info` (pd.DataFrame): DataFrame containing the stable states, cDNFs, and cNDNFs.

    Example:
    **********
    >>> model = blv.Load("model.bnet")
    """
    return BooLEV(file_path)


import time
import re
import itertools
import pandas as pd
from collections import Counter
from pyeda import *
from pyeda.inter import *
from boolevard.utils import *

completePaths = 0
stored_nodes = {}
drivers = {}

def Drivers(model_info, ss):
    '''
    Extracts drivers from (N)DNFs based on the local state of the node within a given stable state.
    
    Arguments
    ---------
    model_info (pd.DataFrame): Info attribute of the BooLEV object.
    ss (int): Stable State to evaluate.
    ---------

    Returns
    ---------
    dict
        Dictionary containing de drivers of each node within the stable state.
    ---------
    '''
    print(f"Evaluating Stable State: {ss}")

    data = pd.concat([model_info[[ss]], model_info[["DNF"]], model_info[["NDNF"]]], axis = 1)
    data_dict = {}

    for node in data.index:
        lsNode = data.loc[node, data.columns[0]]
        stable_state = data.iloc[:, 0]

        if lsNode == 1:
            ruleBlocks = re_notation.sub(lambda x: '~' + x.group(1), model_info.at[node, "DNF"].to_unicode()).replace("·", "&").replace(" ", "").split("+")
        
        else:
            ruleBlocks = re_notation.sub(lambda x: '~' + x.group(1), model_info.at[node, "NDNF"].to_unicode()).replace("·", "&").replace(" ", "").split("+")
        
        blockStates = [expr(re_replace_vars.sub(lambda x: str(stable_state[x.group()]), block)) for block in ruleBlocks]
        driver = "|".join(block for block, state in zip(ruleBlocks, blockStates) if state)
        data_dict[node] = {"state": data.loc[node, ss], "Driver": driver.replace('~','')}
    
    return data_dict
        
def CountPaths(model_info, tNodes, ss_wise = False):
    '''
    Calculates the signed path count leading to the local state of a node contained in a list. Positive if the local state is 1 and negative if 0.
    
    Arguments
    ---------
    model_info (pd.DataFrame): Info attribute of the BooLEV object.
    tNodes (list): list of target nodes to evaluate.
    ss_wise (bool): if True, returns a list with the corresponding path counts leading to a target's local state for each stable state. Otherwise, it computes the average path count across all stable states contained in the Info attribute of the BooLEV object. By default: False.
    ---------

    Returns
    ---------
    list
        Signed number of paths leading to the local states of the target nodes. Negative if the local state is 0, positive if 1.
    ---------
    '''

    # Functions:
    def _LinearPaths(tNode):
        '''
        Calls resolvePaths and asigns a positive sign if the local state of the target node is 1, negative if 0.
        '''
        _ResolvePaths([tNode], True)

        return completePaths if drivers[tNode]["state"] == 1 else -completePaths
    

    def _ResolvePaths(path:list, save):
        '''
        Construct paths leading to the target's node local state and add them to a counter.
        '''
        global completePaths

        path_found_or = False
        blocks = _ExtractNodes(path[-1])
        store_local = {}

        for block in blocks:
            go, path_found_and = _FilterLoops(path, block)
            
            if go:
                path_found_and = True
                
                for elem in block:

                    if elem in store_local:

                        if store_local[elem][1] > 0 and not store_local[elem][0]:
                            path.append(elem)
                            path_found_and_rec = _ResolvePaths(path, save and len(block) == 1)
                            path.pop()

                        else:
                            path_found_and_rec = store_local[elem][0]

                    else:
                        old_len_completePaths = completePaths
                        path.append(elem)
                        path_found_and_rec = _ResolvePaths(path, save and len(block) == 1)
                        path.pop()
                        store_local[elem] = (path_found_and_rec, completePaths - old_len_completePaths)
                    
                    path_found_and = path_found_and and path_found_and_rec
                    
                    if not path_found_and:
                        break

                if path_found_and and save:
                    completePaths += 1
                    path_found_and = False

            path_found_or = path_found_or or path_found_and

        return path_found_or
    
    def _FilterLoops(path, block):
        '''
        Check for loops at current elongation step. Loops will be discarded.
        '''
        if block[0]== path[-1]:
            return False, True
        
        for elem in block:
            if elem in elem in path:
                return False, False
            
        return True, True
    
    def _ExtractNodes(last):
        '''
        Extracts the nodes from a path and replaces them by their drivers.
        '''
        global stored_nodes, drivers
        
        try:
            newBlocks = stored_nodes[last]
        
        except KeyError:
            newBlocks = re_replace_vars.sub(lambda match: f"({drivers[match.group(0)]['Driver']})", last)
            newBlocks = re.sub(r'\(([^)]+)\)', r'(\1)', newBlocks)
        
            if "&" in newBlocks:
                newBlocks = [block.strip('()') for block in re.split(r'\s*&\s*(?=(?:[^()]*\([^()]*\))*[^()]*$)', newBlocks)]
                newBlocks = [block.split("|") for block in newBlocks] 
                newBlocks = ["&".join(pair) for pair in itertools.product(*newBlocks)] 
            
            else:
                newBlocks = newBlocks.replace("(", "").replace(")", "").split("|")
            
            block_lists = [list(set(block.split("&"))) for block in newBlocks]
            total_counter = Counter()
            
            for block in block_lists:
                total_counter.update(block)

            for i in range(len(block_lists)):
                block_lists[i] = sorted(block_lists[i], key=lambda x: total_counter[x], reverse=True)
            
            newBlocks = sorted(block_lists, key=lambda block: max(block, key=lambda x: total_counter[x]), reverse=True)
            stored_nodes[last] = newBlocks

        return newBlocks

    def _ExtractNodesOptimizeDrivers(last): 
        '''
        Retrieves the nodes and processes the node list when drivers are being optimized.
        '''
        global drivers

        newBlocks = re_replace_vars.sub(lambda match: f"({drivers[match.group(0)]['Driver']})", last) 
        newBlocks = re.sub(r'\(([^)]+)\)', r'(\1)', newBlocks) 
        
        if "&" in newBlocks:
            newBlocks = [block.strip('()') for block in re.split(r'\s*&\s*(?=(?:[^()]*\([^()]*\))*[^()]*$)', newBlocks)] 
            newBlocks = [block.split("|") for block in newBlocks] 
            newBlocks = ["&".join(pair) for pair in itertools.product(*newBlocks)] 

        else:
            newBlocks = newBlocks.replace("(", "").replace(")", "").split("|") 
        
        newBlocks = [elem for block in newBlocks for elem in list(set(block.split("&")))]
        
        return newBlocks
    
    def _OptimizeDrivers(drivers):
        '''
        Simplifies the drivers dictionary.
        '''
        for _ in range(6):
            for node in drivers:
                driverNode = _ExtractNodesOptimizeDrivers(node)
                if len(driverNode) == 1:
                    for nodeSearch in drivers:
                        if node in _ExtractNodesOptimizeDrivers(nodeSearch) and nodeSearch != driverNode[0]:
                            drivers[nodeSearch]["Driver"] = re.sub(r'\b'+node+r'\b', driverNode[0], drivers[nodeSearch]["Driver"])
        
        for data_element in drivers:
            drivers[data_element]['Driver'] = drivers[data_element]['Driver'].replace("&", "|")
    
    global completePaths, stored_nodes, drivers
    model_paths = []

    if len(model_info.columns) > 2:
        for state in [state for state in model_info.columns if state not in ["DNF", "NDNF"]]:
            stored_nodes = {}
            drivers = Drivers(model_info, state)
            _OptimizeDrivers(drivers)
            state_paths = []

            for tNode in tNodes:
                completePaths = 0
                start_time = time.time()
                lpaths = _LinearPaths(tNode)
                simulation_time = (time.time() - start_time)/60
                print(f"{tNode}: {lpaths}, {simulation_time} minutes.")
                state_paths.append(lpaths)
            
            model_paths.append(state_paths)

        if ss_wise:
            return model_paths
        
        else:
            model_score = [sum(x)/len(x) for x in zip(*model_paths)]
            model_score = [str(score) for score in model_score]
            return model_score


from setuptools import setup, find_packages

setup(
    name = "boolevard",
    version = "0.1.0",
    description = "BooLEVARD: Boolean Logical Evaluation of Activation and Represion in Directed pathways",
    author = "Marco Fariñas Fernandez",
    author_email = "marco.farinas@gmail.com",
    long_description = open("README.md").read(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/farinasm/boolevard",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages = find_packages(),
    install_requires = [
        "pyeda==0.26.0",
        "mpbn",
        "colomoto_jupyter",
        "pandas"
    ]
)
